#!/usr/bin/python

from daemon import Daemon
from os import sysconf, sysconf_names, statvfs, getcwd
from sys import argv, exit
from re import compile
from time import clock, time, sleep
import json
import urllib2, urllib
from ConfigParser import RawConfigParser
import logging

from socket import setdefaulttimeout, gethostname
from socket import error, gaierror, timeout

logging.basicConfig( filename='/tmp/sb-daemon.log', \
                     format='%(asctime)s : %(message)s', \
                     datefmt='%m%d%y %I:%M:%S %p', level=logging.DEBUG)
log = logging.getLogger('sb-daemon')

log.setLevel(logging.DEBUG)

 
class Checks():
    def __init__(self):
        
        self.AUTHKEY = None
        self.DRY_RUN = None
                     
        self.MONITOR_CPU = None             
        self.MONITOR_NETWORK = None       
        self.MONITOR_LOAD_AVERAGE = None  
        self.MONITOR_MEMORY = None
        self.MONITOR_DISK_USAGE   = None
        self.PARTITIONS  = None
        self.INTERFACES = None
        
        self.memStats = {} ; self.netStats = {} ; self.loadavgStats = {} 
        self.cpuStats = {} ; self.partitionsStats = {}; self.serverStats = {}
        
        setdefaulttimeout(15)
        self._CLOCK_TICKS = sysconf(sysconf_names["SC_CLK_TCK"])

    def getMemStats(self):
        ''' 
        All credit goes to http://forrst.com/posts/python_function_to_read_proc_meminfo-gRj
        '''
        log.debug('getting memory data.')
        re_parser = compile(r'^(?P<key>\S*):\s*(?P<value>\d*)\s*kB')
        result = {}
       
        try:
            for line in open('/proc/meminfo'):
                match = re_parser.match(line)
                if not match:
                    continue 
                key, value = match.groups(['key', 'value'])
                result[key] = int(value)
                
            self.memStats['MemTotal'] = str(result['MemTotal'] / 1024 ) 
            self.memStats['MemFree'] = str( result['MemFree'] / 1024 )
            
            self.memStats['Cached'] = str( result['Cached'] / 1024 )
            
            self.memStats['SwapTotal'] = str( result['SwapTotal'] / 1024 )
            self.memStats['SwapFree'] = str( result['SwapFree'] / 1024 )  
            self.memStats['SwapUsed'] = str( (result['SwapTotal'] - result['SwapFree'] ) / 1024 )
            return self.memStats 
        
        except IOError:
            log.critical('- error while opening /proc/stat. Make sure /proc is mounted.')
            log.critical('- if you don\'t need to monitor CPU disable it in config file.')
            log.critical('- exited.')
            exit(1)

    def getNetworkStats(self):
        '''
        All credit goes to 
        http://stackoverflow.com/questions/1052589/how-can-i-parse-the-output-of-proc-net-dev-into-keyvalue-pairs-per-interface-u
        '''
        log.debug('getting network data.')
        try: 
            lines = open("/proc/net/dev", "r").readlines()
        except IOError, e:
            log.critical('- error while opening /proc/net/dev. Make sure /proc is mounted.')
            log.critical('- if you don\'t need to monitor network disable it in config file.')
            log.critical('- exited.')
            exit(1)
                
        columnLine = lines[1]
        _, receiveCols , transmitCols = columnLine.split("|")
        receiveCols = map(lambda a:"recv_"+a, receiveCols.split())
        transmitCols = map(lambda a:"trans_"+a, transmitCols.split())
        
        cols = receiveCols+transmitCols
        
        faces = {}
        for line in lines[2:]:
            if line.find(":") < 0: continue
            face, data = line.split(":")
            faceData = dict(zip(cols, data.split()))
            faces[face] = faceData

        for eth in faces:
            eths = eth.lstrip()
            if eths in self.INTERFACES:
                self.netStats[eths] = {}
                self.netStats[eths]['recv_bytes'] = str( faces[eth]['recv_bytes'] ) 
                self.netStats[eths]['trans_bytes'] = str( faces[eth]['trans_bytes'] ) 
     
        return self.netStats
    
    def getLoadAvgStats(self):
        
        log.debug('getting load average data.')
        try:
            result =  open("/proc/loadavg").readline().split(" ")[:3]
        
            self.loadavgStats['1']  = str(result[0] )
            self.loadavgStats['5']  = str(result[1] )
            self.loadavgStats['15'] = str(result[2] )
            return self.loadavgStats
        except IOError, e:
            log.critical('- error while opening /proc/net/dev. Make sure /proc is mounted.')
            log.critical('- if you don\'t need to monitor network disable it in config file.')
            log.critical('- exited.')
            exit(1)  
            
    def sendJSON(self, json_dict):
        
        log.debug( 'sending data to serverbeep.com.')
        json_data = json.dumps(json_dict, encoding='latin1').encode('utf-8')
        if self.DRY_RUN:
            
            log.debug("* running in DRY RUN mode. Logging JSON and exiting.")
            log.debug( json_data )
            return None
        keys = json_dict.keys() 
        
        if 'network' not in keys and 'cpu' not in keys and 'diskusage' not in keys and 'loadavg' not in keys:
            log.error('* it seems there\'s no data to send. Make sure sb-daemon is configured properly.')
            return None
         
        headers = {'User-Agent': 'sb-daemon/0.1', \
                   'Content-Type': 'application/x-www-form-urlencoded', \
                   'Accept': 'text/html, */*' }
        
        url = 'http://www.serverbeep.com/servers/push'
        
        try:         
            params = urllib.urlencode({'data' : json_data })
            request = urllib2.Request(url, params, headers)
            response = urllib2.urlopen(request)
            
            http_code = response.getcode()
            body = response.read()
            
            log.debug('response code: {0}, message: {1}'.format(http_code, body))
      
        except Exception, e:
                log.error ( '- exc. caught: {0}'.format(e))
                return None
        return http_code 

    def configParser(self, path):
              
        log.debug('parsing config file {0}.'.format(path))
        
        try:
            config = RawConfigParser()
            config.read(path)
        
            self.AUTHKEY = config.get('Main', 'AUTHKEY')
            self.DRY_RUN =  config.getint('Main', 'DRY_RUN')
            self.MONITOR_CPU = config.getint('Main', 'MONITOR_CPU')             
            self.MONITOR_NETWORK = config.getint('Main', 'MONITOR_NETWORK')       
            self.MONITOR_LOAD_AVERAGE = config.getint('Main', 'MONITOR_LOAD_AVERAGE')  
            self.MONITOR_MEMORY = config.getint('Main', 'MONITOR_MEMORY')
            self.MONITOR_DISK_USAGE   = config.getint('Main', 'MONITOR_DISK_USAGE')
               
            log.debug('monitoring items. cpu: {0}, network: {1}, load average: {2}, memory: {3}, disk usage: {4}.'.format(self.MONITOR_CPU, \
            self.MONITOR_NETWORK, self.MONITOR_LOAD_AVERAGE,  self.MONITOR_MEMORY, self.MONITOR_DISK_USAGE ))                            
            self.PARTITIONS = config.get('Main', 'PARTITIONS').split(",")
            self.INTERFACES = config.get('Main', 'NETWORK_INTERFACES').split(",")
            
        except Exception, e:
            log.critical('- error while parsing config file: \"{0}\".'.format(e))
            log.critical('- make sure you have sb-daemon.cfg file and it has valid format.')
            log.critical('- exited.')
            exit(1)
            
        if self.DRY_RUN:
        
            log.debug('* running in DRY RUN mode.')
         
        self.serverStats['authkey'] = self.AUTHKEY
        if not self.AUTHKEY:
            log.critical('- there\'s  no authkey. please add your authkey to config.')
            log.critical('- exited.')
            exit(1)                       
        
        return self    

    def getSystemCpuTimes(self):
        """
        All credit goes to http://code.google.com/p/psutil/
        
        Return a _list_ representing the following CPU times:
        user, nice, system, idle, iowait, irq, softirq.
        """
        
        try:
            f = open('/proc/stat', 'r')
            values = f.readline().split()
            values = values[1:8]
            values = tuple([float(x) / self._CLOCK_TICKS for x in values])
            f.close()
            return values[:7]
        
        except IOError:
            log.critical('- error while opening /proc/stat. Make sure /proc is mounted.')
            log.critical('- if you don\'t need to monitor CPU disable it in config file.')
            log.critical('- exited.')
            exit(1)
                 
    def getCpuStats(self):
        
        log.debug('getting CPU data.')
        result = {}
        time1 = 0; time2 = 0
        times1 =  self.getSystemCpuTimes()
        for i in times1:
            time1 = time1 + i
        sleep(30)
        times2 = self.getSystemCpuTimes()
        for i in times2:
            time2 = time2 + i
        time_total = 0; time_total = time2 - time1
        
        self.cpuStats['cpu_user']    = "%.2f" %((( times2[0] - times1[0] ) / time_total ) * 100 )
        self.cpuStats['cpu_nice']    = "%.2f" %((( times2[1] - times1[1] ) / time_total ) * 100 )
        self.cpuStats['cpu_system']  = "%.2f" %((( times2[2] - times1[2] ) / time_total ) * 100 )
        self.cpuStats['cpu_idle']    = "%.2f" %((( times2[3] - times1[3] ) / time_total ) * 100 )
        self.cpuStats['cpu_irq']     = "%.2f" %((( times2[4] - times1[4] ) / time_total ) * 100 )
        self.cpuStats['cpu_iowait']  = "%.2f" %((( times2[5] - times1[5] ) / time_total ) * 100 )
        self.cpuStats['cpu_softirq'] = "%.2f" %((( times2[6] - times1[6] ) / time_total ) * 100 )
        
        return self.cpuStats
    
    def getDiskUsage(self):
        """ Return disk usage associated with path."""
        
        log.debug('getting partitions data.')
        for partition in self.PARTITIONS:
            
            try:
                st = statvfs(partition)
            except OSError, e:
                log.critical(' -error while checking disk usage: "{0}".'.format(e))
                log.critical(' -make sure you configured the partitions properly. exiting.')
                log.critical(' -if you don\'t need to monitor disk usage disable it in config file.')
                log.critical('exited.')
                exit(1)
                
            free = (st.f_bavail * st.f_frsize) / 1024 / 1024
            total = (st.f_blocks * st.f_frsize) / 1024 / 1024
            used = ((st.f_blocks - st.f_bfree) * st.f_frsize ) / 1024 / 1024
            
            self.partitionsStats[partition] = {}
            self.partitionsStats[partition]['free'] = free
            self.partitionsStats[partition]['total'] = total
            self.partitionsStats[partition]['used'] = used
   
        return self.partitionsStats

class Yeehaa(Daemon):

    def run(self):
        log.debug('+ sb-daemon started.')
        Z = Checks()
        Z.configParser('/home/serverbeep/sb-daemon/sb-daemon.cfg')
        
        while True:
            
            t1 = clock()
            p1 = time()   
            
            if Z.MONITOR_MEMORY:
                Z.getMemStats()
                log.debug( Z.memStats)
                Z.serverStats['memory'] = Z.memStats
                
            if Z.MONITOR_NETWORK:    
                Z.getNetworkStats()
                log.debug( Z.netStats)
                Z.serverStats['network'] = Z.netStats
            
            if Z.MONITOR_LOAD_AVERAGE:
                Z.getLoadAvgStats()
                log.debug ( Z.loadavgStats )
                Z.serverStats['loadavg'] = Z.loadavgStats
                
            if Z.MONITOR_CPU:
               Z.getCpuStats()
               log.debug ( Z.cpuStats )
               Z.serverStats['cpu'] = Z.cpuStats
               
            if Z.MONITOR_DISK_USAGE:
                Z.getDiskUsage()
                log.debug ( Z.partitionsStats )
                Z.serverStats['diskusage'] = Z.partitionsStats
       
            if not Z.DRY_RUN:
                
                log.debug( '*** summary ***')
                log.debug( Z.serverStats)
            
            Z.sendJSON(Z.serverStats)
            
            t2 = clock()
            p2 = time()
            delta = t2-t1
            pelta = p2-p1
                
            log.critical("time taken system: {0} seconds.".format( delta ) )
            log.critical("time taken real: {0} seconds.".format( int( pelta ) ))
            
            sweet_dream = int( 299 - int(pelta) )
            log.critical("let's dream for: {0} seconds.".format( sweet_dream ) )
            sleep(sweet_dream)
            log.debug('+ running next cycle.')
            log.debug('')
     
if __name__ == "__main__":

        daemon = Yeehaa('/tmp/sb-daemon.pid')
        
        if len(argv) == 2:
                if 'start' == argv[1]:
                        daemon.start()
                elif 'stop' == argv[1]:
                        daemon.stop()
                elif 'restart' == argv[1]:
                        daemon.restart()
                else:
                        print "Unknown command"
                        exit(2)
                exit(0)
        else:
                print "usage: %s start|stop|restart" % argv[0]
                exit(2)   
                           


